 #ifndef _APP_H_
#define _APP_H_

typedef unsigned char UINT8;
typedef unsigned short UINT16;
typedef unsigned int UINT32;
typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint32;

typedef signed char INT8;
typedef signed short INT16;
typedef signed int INT32;

typedef signed char int8;
typedef signed short int16;
typedef signed int int32;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include <fileio.h>

#include "../Cyclone/Cyclone.h"
#include "./m68k/m68k.h"
#include "../DrZ80/DrZ80.h"
#include "DrMD.h"
#include "DrSMS.h"
#include "fm.h"
#include "sn76496.h"
#include "unzip.h"
#include "zip.h"
#include "savestate.h"


#ifdef __GP32__
#include "gp32_snd.h"
#endif

#ifdef __GP2X__
#include <linux/errno.h>
#include <dirent.h>
#include <errno.h>
#include "gp2x_sdk.h"
#endif

#ifdef __GIZ__
#include <dirent.h>
#include <errno.h>

#include <sys/wcetypes.h>
#ifdef __SDL__
#include <SDL.H>
#include "giz_sdk_sdl.h"
#else
#include <KGSDK.h>
#include <Framework.h>
#include <Framework2D.h>
#include "giz_sdk_kgsdk.h"
#endif


#endif

#define EMU_C68K
//#define EMU_M68K




#ifdef __GP32__ 
#define DIR_SEPERATOR			"\\"
#define ROM_PATH_GG 	"dev0:\\GPMM\\ROMS\\GG"
#define ROM_PATH_SMS 	"dev0:\\GPMM\\ROMS\\SMS"
#define ROM_PATH_MD 	"dev0:\\GPMM\\ROMS\\MD"

#define SYSTEM_DIR		"dev0:\\GPMM\\DRMD"
#define SMS_SYSTEM_DIR		"dev0:\\GPMM\\DRMD\\SMS"
#define GG_SYSTEM_DIR		"dev0:\\GPMM\\DRMD\\GG"
#define MD_SYSTEM_DIR		"dev0:\\GPMM\\DRMD\\MD"
#endif

#ifdef __GP2X__
#define DIR_SEPERATOR			"/"
#define ROM_PATH_GG 	"/mnt/sd/roms/gg"
#define ROM_PATH_SMS 	"/mnt/sd/roms/sms"
#define ROM_PATH_MD 	"/mnt/sd/roms/megadrive"

#define SYSTEM_DIR		"/mnt/sd/drmd"
#define SMS_SYSTEM_DIR		"sms"
#define GG_SYSTEM_DIR		"gg"
#define MD_SYSTEM_DIR		"md"
#endif

#ifdef __GIZ__ 
#define DIR_SEPERATOR			"\\"
#define ROM_PATH_GG 	"\\SD Card\\DrMD\\ROMS\\GG"
#define ROM_PATH_SMS 	"\\SD Card\\DrMD\\ROMS\\SMS"
#define ROM_PATH_MD 	"\\SD Card\\DrMD\\ROMS\\MD"

#define SYSTEM_DIR		"\\SD Card\\DrMD"
#define SMS_SYSTEM_DIR		"SMS"
#define GG_SYSTEM_DIR		"GG"
#define MD_SYSTEM_DIR		"MD"
#endif

#define ROM_LIST_FILENAME			"romlist.bin"
#define SRAM_FILE_EXT					"srm"
#define SAVESTATE_EXT					"sv"
#define MENU_OPTIONS_FILENAME		"menu"
#define MENU_OPTIONS_EXT				"opt"
#define DEFAULT_ROM_DIR_FILENAME		"romdir"
#define DEFAULT_ROM_DIR_EXT			"opt"


//define emulation modes
#define EMU_MODE_NONE	0
#define EMU_MODE_MD 	1
#define EMU_MODE_SMS	2
#define EMU_MODE_GG		3

#define SAVESTATE_MODE_SAVE				0
#define SAVESTATE_MODE_LOAD				1
#define SAVESTATE_MODE_DELETE			2

#define MD_CLOCK_NTSC 				53693175.0
#define MD_CLOCK_PAL  				53203424.0
#define MD_OPTIONS_VER 			6
#define SMS_OPTIONS_VER 			6
#define GG_OPTIONS_VER 			6
#define DRMD_VERSION				"V5 beta 7"

#ifdef __GP32__
#define ROM_SIZE 0x400000 //most games
//#define ROM_SIZE 				0x500000 //ssf2(40mbits)
#define RGB(r,g,b) ((r) << 11 | (g) << 6 | (b) << 1 ) 
#define MAX_ROMS		512
#define MAX_CPU		39

#define MENU_CPU_SPEED 			40
#define MENU_FAST_CPU_SPEED		133
#endif

#ifdef __GP2X__
#define ROM_SIZE 				0x500000 //ssf2(40mbits)
#define RGB(r,g,b) ((r) << 11 | (g) << 6 | (b) << 0 )
#define MAX_ROMS		3000
#define MAX_CPU		39

#define MENU_CPU_SPEED 			80
#define MENU_FAST_CPU_SPEED		200
#endif

#ifdef __GIZ__
#define ROM_SIZE 				0x500000 //ssf2(40mbits)
#define RGB(r,g,b) ((r) << 11 | (g) << 6 | (b) << 0 )
#define MAX_ROMS		3000
#define MAX_CPU		39

#define MENU_CPU_SPEED 			80
#define MENU_FAST_CPU_SPEED		200
#endif

#define MAX_PATH    255

#define FILE_TYPE_FILE									0
#define FILE_TYPE_DIRECTORY							1

#ifdef __GP32__
#if defined (__EMU_MD__)
#define MAIN_MENU_RETURN								0
#define MAIN_MENU_ROM_SELECT						1
#define MAIN_MENU_MANAGE_QUICK_STATE			2
#define MAIN_MENU_MANAGE_SAVE_STATE			3
#define MAIN_MENU_MANAGE_SRAM						4
#define MAIN_MENU_MD_OPTIONS						5
#define MAIN_MENU_RESET_GAME						6
#define MAIN_MENU_LCD_TYPE               	7
#define MAIN_MENU_EXIT_APP							8
#define MAIN_MENU_COUNT								9
#endif
#if defined (__EMU_SMS__)
#define MAIN_MENU_RETURN								0
#define MAIN_MENU_ROM_SELECT						1
#define MAIN_MENU_MANAGE_QUICK_STATE			2
#define MAIN_MENU_MANAGE_SAVE_STATE			3
#define MAIN_MENU_MANAGE_SRAM						4
#define MAIN_MENU_SMS_OPTIONS						5
#define MAIN_MENU_GG_OPTIONS						6
#define MAIN_MENU_RESET_GAME						7
#define MAIN_MENU_LCD_TYPE               	8
#define MAIN_MENU_EXIT_APP							9
#define MAIN_MENU_COUNT								10
#endif
#endif

#ifdef __GP2X__
#define MAIN_MENU_RETURN								0
#define MAIN_MENU_ROM_SELECT						1
#define MAIN_MENU_MANAGE_QUICK_STATE			2
#define MAIN_MENU_MANAGE_SAVE_STATE			3
#define MAIN_MENU_MANAGE_SRAM						4
#define MAIN_MENU_MD_OPTIONS						5
#define MAIN_MENU_SMS_OPTIONS						6
#define MAIN_MENU_GG_OPTIONS						7
#define MAIN_MENU_RESET_GAME						8
#define MAIN_MENU_EXIT_APP							9
#define MAIN_MENU_COUNT								10
#endif

#ifdef __GIZ__
#define MAIN_MENU_RETURN								0
#define MAIN_MENU_ROM_SELECT						1
#define MAIN_MENU_MANAGE_QUICK_STATE			2
#define MAIN_MENU_MANAGE_SAVE_STATE			3
#define MAIN_MENU_MANAGE_SRAM						4
#define MAIN_MENU_MD_OPTIONS						5
#define MAIN_MENU_SMS_OPTIONS						6
#define MAIN_MENU_GG_OPTIONS						7
#define MAIN_MENU_RESET_GAME						8
#define MAIN_MENU_EXIT_APP							9
#define MAIN_MENU_COUNT								10
#endif

#ifdef __GP32__
#if defined (__EMU_MD__)
#define LOAD_ROM_MENU_MD								0
#define LOAD_ROM_MENU_RETURN						1
#define LOAD_ROM_MENU_COUNT							2
#endif
#if defined (__EMU_SMS__)
#define LOAD_ROM_MENU_SMS							0
#define LOAD_ROM_MENU_GG								1
#define LOAD_ROM_MENU_RETURN						2
#define LOAD_ROM_MENU_COUNT							3
#endif
#endif

#if defined (__GP2X__) || defined (__GIZ__)
#define LOAD_ROM_MENU_MD								0
#define LOAD_ROM_MENU_SMS							1
#define LOAD_ROM_MENU_GG								2
#define LOAD_ROM_MENU_RETURN						3
#define LOAD_ROM_MENU_COUNT							4
#endif

#ifdef __GP32__
#define MD_MENU_SOUND									0
#define MD_MENU_SOUND_RATE              	1
#define MD_MENU_FRAMESKIP               	2
#define MD_MENU_CPU                     	3
#define MD_MENU_CONTROLS                	4
#define MD_MENU_PAD_TYPE								5
#define MD_MENU_REGION                  	6
#define MD_MENU_FPS                     	7
#define MD_MENU_GAMMA                   	8
#define MD_MENU_AUTO_SRAM							9
#define MD_MENU_LOAD_GLOBAL							10
#define MD_MENU_SAVE_GLOBAL							11
#define MD_MENU_DELETE_GLOBAL						12
#define MD_MENU_LOAD_CURRENT						13
#define MD_MENU_SAVE_CURRENT						14
#define MD_MENU_DELETE_CURRENT					15
#define MD_MENU_SET_ROMDIR							16
#define MD_MENU_RETURN									17
#define MD_MENU_COUNT									18

#define SMS_MENU_SOUND									0
#define SMS_MENU_SOUND_RATE              	1
#define SMS_MENU_FRAMESKIP               	2
#define SMS_MENU_RENDER_MODE						3
#define SMS_MENU_CPU                     	4
#define SMS_MENU_CONTROLS                	5
#define SMS_MENU_REGION                  	6
#define SMS_MENU_FPS                     	7
#define SMS_MENU_GAMMA                   	8
#define SMS_MENU_AUTO_SRAM							9
#define SMS_MENU_LOAD_GLOBAL						10
#define SMS_MENU_SAVE_GLOBAL						11
#define SMS_MENU_DELETE_GLOBAL					12
#define SMS_MENU_LOAD_CURRENT						13
#define SMS_MENU_SAVE_CURRENT						14
#define SMS_MENU_DELETE_CURRENT					15
#define SMS_MENU_SET_ROMDIR							16
#define SMS_MENU_RETURN								17
#define SMS_MENU_COUNT									18

#define GG_MENU_SOUND									0
#define GG_MENU_SOUND_RATE              	1
#define GG_MENU_FRAMESKIP               	2
#define GG_MENU_RENDER_MODE							3
#define GG_MENU_CPU                     	4
#define GG_MENU_CONTROLS                	5
#define GG_MENU_REGION                  	6
#define GG_MENU_FPS                     	7
#define GG_MENU_GAMMA                   	8
#define GG_MENU_AUTO_SRAM							9
#define GG_MENU_LOAD_GLOBAL							10
#define GG_MENU_SAVE_GLOBAL							11
#define GG_MENU_DELETE_GLOBAL						12
#define GG_MENU_LOAD_CURRENT						13
#define GG_MENU_SAVE_CURRENT						14
#define GG_MENU_DELETE_CURRENT					15
#define GG_MENU_SET_ROMDIR							16
#define GG_MENU_RETURN									17
#define GG_MENU_COUNT									18
#endif

#ifdef __GP2X__
#define MD_MENU_SOUND									0
#define MD_MENU_SOUND_RATE              	1
#define MD_MENU_SOUND_VOL				       	2
#define MD_MENU_FRAMESKIP               	3
#define MD_MENU_CPU                     	4
#define MD_MENU_CONTROLS                	5
#define MD_MENU_PAD_TYPE								6
#define MD_MENU_REGION                  	7
#define MD_MENU_FPS                     	8
#define MD_MENU_GAMMA                   	9
#define MD_MENU_AUTO_SRAM							10
#define MD_MENU_LOAD_GLOBAL							11
#define MD_MENU_SAVE_GLOBAL							12
#define MD_MENU_DELETE_GLOBAL						13
#define MD_MENU_LOAD_CURRENT						14
#define MD_MENU_SAVE_CURRENT						15
#define MD_MENU_DELETE_CURRENT					16
#define MD_MENU_SET_ROMDIR							17
#define MD_MENU_RETURN									18
#define MD_MENU_COUNT									19

#define SMS_MENU_SOUND									0
#define SMS_MENU_SOUND_RATE              	1
#define SMS_MENU_SOUND_VOL              	2
#define SMS_MENU_FRAMESKIP               	3
#define SMS_MENU_RENDER_MODE						4
#define SMS_MENU_CPU                     	5
#define SMS_MENU_CONTROLS                	6
#define SMS_MENU_REGION                  	7
#define SMS_MENU_FPS                     	8
#define SMS_MENU_GAMMA                   	9
#define SMS_MENU_AUTO_SRAM							10
#define SMS_MENU_LOAD_GLOBAL						11
#define SMS_MENU_SAVE_GLOBAL						12
#define SMS_MENU_DELETE_GLOBAL					13
#define SMS_MENU_LOAD_CURRENT						14
#define SMS_MENU_SAVE_CURRENT						15
#define SMS_MENU_DELETE_CURRENT					16
#define SMS_MENU_SET_ROMDIR							17
#define SMS_MENU_RETURN								18
#define SMS_MENU_COUNT									19

#define GG_MENU_SOUND									0
#define GG_MENU_SOUND_RATE              	1
#define GG_MENU_SOUND_VOL	              	2
#define GG_MENU_FRAMESKIP               	3
#define GG_MENU_RENDER_MODE							4
#define GG_MENU_CPU                     	5
#define GG_MENU_CONTROLS                	6
#define GG_MENU_REGION                  	7
#define GG_MENU_FPS                     	8
#define GG_MENU_GAMMA                   	9
#define GG_MENU_AUTO_SRAM							10
#define GG_MENU_LOAD_GLOBAL							11
#define GG_MENU_SAVE_GLOBAL							12
#define GG_MENU_DELETE_GLOBAL						13
#define GG_MENU_LOAD_CURRENT						14
#define GG_MENU_SAVE_CURRENT						15
#define GG_MENU_DELETE_CURRENT					16
#define GG_MENU_SET_ROMDIR							17
#define GG_MENU_RETURN									18
#define GG_MENU_COUNT									19
#endif

#ifdef __GIZ__
#define MD_MENU_SOUND									0
#define MD_MENU_SOUND_RATE              	1
#define MD_MENU_SOUND_VOL				       	2
#define MD_MENU_FRAMESKIP               	3
#define MD_MENU_CPU                     	4
#define MD_MENU_CONTROLS                	5
#define MD_MENU_PAD_TYPE								6
#define MD_MENU_REGION                  	7
#define MD_MENU_FPS                     	8
#define MD_MENU_GAMMA                   	9
#define MD_MENU_AUTO_SRAM							10
#define MD_MENU_LOAD_GLOBAL							11
#define MD_MENU_SAVE_GLOBAL							12
#define MD_MENU_DELETE_GLOBAL						13
#define MD_MENU_LOAD_CURRENT						14
#define MD_MENU_SAVE_CURRENT						15
#define MD_MENU_DELETE_CURRENT					16
#define MD_MENU_SET_ROMDIR							17
#define MD_MENU_RETURN									18
#define MD_MENU_COUNT									19

#define SMS_MENU_SOUND									0
#define SMS_MENU_SOUND_RATE              	1
#define SMS_MENU_SOUND_VOL              	2
#define SMS_MENU_FRAMESKIP               	3
#define SMS_MENU_RENDER_MODE						4
#define SMS_MENU_CPU                     	5
#define SMS_MENU_CONTROLS                	6
#define SMS_MENU_REGION                  	7
#define SMS_MENU_FPS                     	8
#define SMS_MENU_GAMMA                   	9
#define SMS_MENU_AUTO_SRAM							10
#define SMS_MENU_LOAD_GLOBAL						11
#define SMS_MENU_SAVE_GLOBAL						12
#define SMS_MENU_DELETE_GLOBAL					13
#define SMS_MENU_LOAD_CURRENT						14
#define SMS_MENU_SAVE_CURRENT						15
#define SMS_MENU_DELETE_CURRENT					16
#define SMS_MENU_SET_ROMDIR							17
#define SMS_MENU_RETURN								18
#define SMS_MENU_COUNT									19

#define GG_MENU_SOUND									0
#define GG_MENU_SOUND_RATE              	1
#define GG_MENU_SOUND_VOL	              	2
#define GG_MENU_FRAMESKIP               	3
#define GG_MENU_RENDER_MODE							4
#define GG_MENU_CPU                     	5
#define GG_MENU_CONTROLS                	6
#define GG_MENU_REGION                  	7
#define GG_MENU_FPS                     	8
#define GG_MENU_GAMMA                   	9
#define GG_MENU_AUTO_SRAM							10
#define GG_MENU_LOAD_GLOBAL							11
#define GG_MENU_SAVE_GLOBAL							12
#define GG_MENU_DELETE_GLOBAL						13
#define GG_MENU_LOAD_CURRENT						14
#define GG_MENU_SAVE_CURRENT						15
#define GG_MENU_DELETE_CURRENT					16
#define GG_MENU_SET_ROMDIR							17
#define GG_MENU_RETURN									18
#define GG_MENU_COUNT									19
#endif



#define QUICKSTATE_MENU_LOAD						0
#define QUICKSTATE_MENU_SAVE						1
#define QUICKSTATE_MENU_RETURN					2
#define QUICKSTATE_MENU_COUNT						3

#define SAVESTATE_MENU_LOAD						0
#define SAVESTATE_MENU_SAVE						1
#define SAVESTATE_MENU_DELETE					2
#define SAVESTATE_MENU_RETURN					3
#define SAVESTATE_MENU_COUNT					4

#define SRAM_MENU_LOAD						0
#define SRAM_MENU_SAVE						1
#define SRAM_MENU_DELETE					2
#define SRAM_MENU_RETURN					3
#define SRAM_MENU_COUNT					4

#define EVENT_EXIT_APP						1
#define EVENT_LOAD_MD_ROM				2
#define EVENT_RUN_MD_ROM					3
#define EVENT_LOAD_SMS_ROM				4
#define EVENT_RUN_SMS_ROM				5
#define EVENT_LOAD_GG_ROM				6
#define EVENT_RUN_GG_ROM					7

#define menutile_width      64
#define menutile_height     64

#define GP32_GCC

   /*
 MD PAD
   Convert to Megadrive pad: format is SACB RLDU
   0x01 = UP
   0x02 = DOWN
   0x04 = LEFT
   0x08 = RIGHT
   0x10 = B
   0x20 = C
   0x40 = A
   0x80 = START
 
 */
#define MD_BUTTON_UP     0
#define MD_BUTTON_DOWN   1 
#define MD_BUTTON_LEFT   2  
#define MD_BUTTON_RIGHT  3 
#define MD_BUTTON_B      4
#define MD_BUTTON_C      5 
#define MD_BUTTON_A      6 
#define MD_BUTTON_START  7
#define MD_BUTTON_Z  		8
#define MD_BUTTON_Y  		9
#define MD_BUTTON_X  		10
#define MD_BUTTON_MODE  	11 

#define SMS_BUTTON_UP     0
#define SMS_BUTTON_DOWN   1 
#define SMS_BUTTON_LEFT   2 
#define SMS_BUTTON_RIGHT  3 
#define SMS_BUTTON_1      4 
#define SMS_BUTTON_2      5 
#define SMS_BUTTON_START  7 

#define GG_BUTTON_UP     0
#define GG_BUTTON_DOWN   1 
#define GG_BUTTON_LEFT   2  
#define GG_BUTTON_RIGHT  3
#define GG_BUTTON_1      4 
#define GG_BUTTON_2      5 
#define GG_BUTTON_START  7 

#ifdef __GP32__
#define INP_BUTTON_LEFT     0
#define INP_BUTTON_DOWN     1
#define INP_BUTTON_RIGHT    2
#define INP_BUTTON_UP       3
#define INP_BUTTON_L        4
#define INP_BUTTON_B        5
#define INP_BUTTON_A        6
#define INP_BUTTON_R        7
#define INP_BUTTON_START    8
#define INP_BUTTON_SELECT   9
#define INP_BUTTON_MENU_SELECT	INP_BUTTON_A
#define INP_BUTTON_MENU_CANCEL	INP_BUTTON_B
#define INP_BUTTON_MENU_DELETE	INP_BUTTON_SELECT
#endif

#ifdef __GP2X__
#define INP_BUTTON_MENU_SELECT	INP_BUTTON_B
#define INP_BUTTON_MENU_CANCEL	INP_BUTTON_X
#define INP_BUTTON_MENU_ENTER		INP_BUTTON_SELECT
#define INP_BUTTON_MENU_DELETE		INP_BUTTON_SELECT
#define INP_BUTTON_MENU_QUICKSAVE1		INP_BUTTON_R
#define INP_BUTTON_MENU_QUICKSAVE2		INP_BUTTON_SELECT
#define INP_BUTTON_MENU_QUICKLOAD1		INP_BUTTON_L
#define INP_BUTTON_MENU_QUICKLOAD2		INP_BUTTON_SELECT
#endif

#ifdef __GIZ__
#define INP_BUTTON_MENU_SELECT	INP_BUTTON_PLAY
#define INP_BUTTON_MENU_CANCEL	INP_BUTTON_STOP
#define INP_BUTTON_MENU_ENTER		INP_BUTTON_BRIGHT
#define INP_BUTTON_MENU_DELETE		INP_BUTTON_REWIND
#define INP_BUTTON_MENU_QUICKSAVE1		INP_BUTTON_R
#define INP_BUTTON_MENU_QUICKSAVE2		INP_BUTTON_BRIGHT
#define INP_BUTTON_MENU_QUICKLOAD1		INP_BUTTON_L
#define INP_BUTTON_MENU_QUICKLOAD2		INP_BUTTON_BRIGHT
#endif

typedef struct {
   char name[MAX_ROMS][MAX_PATH];    //  128 entrys,16 Bytes long
   int size[MAX_ROMS];
} DIRDATA;

//Graphics - moved to objects because they get updated with current gamma setting
extern unsigned short menu_header[];
extern unsigned short menu_header_orig[];
extern unsigned short highlightbar[];
extern unsigned short highlightbar_orig[];
extern unsigned short menutile[];
extern unsigned short menutile_orig[];

// Main.cpp
#if 0
#ifdef __GIZ__
extern SDL_Joystick *joystick;
#endif
#endif

extern struct DrMD drmd;
extern struct DrSMS drsms;
extern struct DrZ80 drz80;
#ifdef EMU_C68K
extern struct Cyclone cyclone;
#endif

extern unsigned char pad_config[];
extern float sound_rates[];
extern char system_dir[];
extern char system_root_dir[];
extern char md_system_dir[];
extern char sms_system_dir[];
extern char gg_system_dir[];
extern unsigned char gamma_conv[];
extern unsigned short sample_count_lookup[];
extern int sound_stage_lookup[];
extern int fm_timera_tab[];   /* Precalculated timer A values */
extern int fm_timerb_tab[];   /* Precalculated timer B values */
extern unsigned char work_ram[]; // scratch ram
extern unsigned char sram[];
extern unsigned short vram[];
extern unsigned char zram[]; // Z80 ram
extern unsigned short cram[];
extern unsigned short vsram[];
extern unsigned int pal_lookup[];
extern char last_save_name[];
extern short *soundbuffer;
extern unsigned char *RomData;
extern unsigned char* framebuffer8[];
extern unsigned short* framebuffer16[];
extern int CurrentEmuMode;
extern int gp32_lcdver;
extern int laststage;
extern int Flip;
extern int prevFlip;
extern int savestatesize;
extern int romloaded;
extern int Frames,Taken; // Frames and 60hz ticks
extern unsigned int testTimer;
extern unsigned int testTimerDiff;
extern unsigned int testLen;
extern char show_fps;
extern char sound_rate;
extern char sound_on;

#ifdef EMU_M68K
int vdp_int_ack_callback(int int_level);
#endif
void Convert8bitTo16bit(int line);
void update_menu_graphics_gamma(void);
void MD_load_pal(void);
void SMS_load_pal(void);
int Round_Double(double val);
void MD_update_sound_timing(void);
void SMS_update_sound_timing(void);
void clear_screen(unsigned int *buffer,unsigned int data);
void InvalidOpCallback(unsigned int opcode);
void reset_drmd(void);
int SekInterrupt(int irq);
unsigned int DrMDCheckPc(unsigned int pc);
void Cyclone_Reset();
void Cyclone_Init();
void Cyclone_ClearIRQ(int irq);
void Genesis_Init(void);
void Detect_Country_Genesis(void);
void DrMD_Reset();
void DrMD_Init();
void DrSMS_Reset();
void DrSMS_Init();
int CheckSram();
void DrMD_Fix_Controllers(void);
void LoadSram(char *path,char *romname,char *ext,char *srammem);
void SaveSram(char *path,char *romname,char *ext,char *srammem);
void DeleteSram(char *path,char *romname,char *ext);
int SaveMenuOptions(char *path, char *filename, char *ext, char *optionsmem, int maxsize);
int LoadMenuOptions(char *path, char *filename, char *ext, char *optionsmem, int maxsize);
int DeleteMenuOptions(char *path, char *filename, char *ext);

#if defined (__GP32__) || defined (__GIZ__)
void sync(void);
#endif

void DrMD_DrZ80_Reset();
void DrMD_DrZ80_Init();
void DrMD_DrZ80_Set_Irq(unsigned int irq);
void DrMD_DrZ80_irq_callback(void);

void DrSMS_DrZ80_Reset();
void DrSMS_DrZ80_Init();
void DrSMS_DrZ80_Set_Irq(unsigned int irq);
void DrSMS_DrZ80_irq_callback(void);

// menu.cpp
void MenuPause(void);
void MenuFlip(void);
void SplitFilename(char *wholefilename, char *filename, char *ext);
int FileSelect(int mode, char *romlistname, int rom_type);
int MainMenu();
void PrintTitle(int flip);
void PrintTile(int flip);
void PrintBar(int flip, unsigned int given_y);

int load_rom_list(char *romlistname, int rom_type);
int FileScan(char *romlistname,int rom_type);
extern void loadstate_file(char *filename);

extern int rom_list_loaded;
extern int quick_save_present;
extern unsigned short cpu_speed_lookup[];
extern float gamma_lookup[];

struct RomList_Item
{
  unsigned int type;
  unsigned int crc;
  unsigned int filesize;
  unsigned int romsize;
  char shortname[MAX_PATH+1];
  char longname[MAX_PATH+1];
};

extern struct RomList_Item romlist[];
extern int currentrom;
extern char currentrom_shortname[];
extern char RomDir[];
extern char MD_RomDir[];
extern char SMS_RomDir[];
extern char GG_RomDir[];

struct MD_Menu_Options
{
  unsigned char menu_ver;
  unsigned char frameskip;
  unsigned char sound_on;
  unsigned char cpu_speed;
  unsigned char pad_config[32];
  unsigned char triple_buffer;
  unsigned char force_region;
  unsigned char show_fps;
  unsigned char gamma;
  unsigned char lcdver;
  unsigned char stereo;
  unsigned char sound_rate;
  unsigned char auto_sram;
  unsigned char render_mode;
  unsigned char volume;
  unsigned char pad_type;
  unsigned char spare10;
  unsigned char spare11;
  unsigned char spare12;
  unsigned char spare13;
  unsigned char spare14;
  unsigned char spare15;
  unsigned char spare16;
  unsigned char spare17;
  unsigned char spare18;
  unsigned char spare19;
  unsigned char spare1A;
  unsigned char spare1B;
  unsigned char spare1C;
  unsigned char spare1D;
  unsigned char spare1E;
  unsigned char spare1F;
};

extern struct MD_Menu_Options md_menu_options;

struct SMS_Menu_Options
{
  unsigned char menu_ver;
  unsigned char frameskip;
  unsigned char sound_on;
  unsigned char cpu_speed;
  unsigned char pad_config[32];
  unsigned char triple_buffer;
  unsigned char force_region;
  unsigned char show_fps;
  unsigned char gamma;
  unsigned char lcdver;
  unsigned char stereo;
  unsigned char sound_rate;
  unsigned char auto_sram;
  unsigned char render_mode;
  unsigned char volume;
  unsigned char spare0F;
  unsigned char spare10;
  unsigned char spare11;
  unsigned char spare12;
  unsigned char spare13;
  unsigned char spare14;
  unsigned char spare15;
  unsigned char spare16;
  unsigned char spare17;
  unsigned char spare18;
  unsigned char spare19;
  unsigned char spare1A;
  unsigned char spare1B;
  unsigned char spare1C;
  unsigned char spare1D;
  unsigned char spare1E;
  unsigned char spare1F;
};

extern struct SMS_Menu_Options sms_menu_options;

struct GG_Menu_Options
{
  unsigned char menu_ver;
  unsigned char frameskip;
  unsigned char sound_on;
  unsigned char cpu_speed;
  unsigned char pad_config[32];
  unsigned char triple_buffer;
  unsigned char force_region;
  unsigned char show_fps;
  unsigned char gamma;
  unsigned char lcdver;
  unsigned char stereo;
  unsigned char sound_rate;
  unsigned char auto_sram;
  unsigned char render_mode;
  unsigned char volume;
  unsigned char spare0F;
  unsigned char spare10;
  unsigned char spare11;
  unsigned char spare12;
  unsigned char spare13;
  unsigned char spare14;
  unsigned char spare15;
  unsigned char spare16;
  unsigned char spare17;
  unsigned char spare18;
  unsigned char spare19;
  unsigned char spare1A;
  unsigned char spare1B;
  unsigned char spare1C;
  unsigned char spare1D;
  unsigned char spare1E;
  unsigned char spare1F;
};

extern struct GG_Menu_Options gg_menu_options;

struct SaveState
{
  char filename[MAX_PATH+1];
  char fullfilename[MAX_PATH+1];
  unsigned int inuse;
};


extern struct SaveState savestate[];  // holds the filenames for the savestate and "inuse" flags
extern char savestate_name[];

// Input.cpp
struct Input
{
  unsigned int held[32];
  unsigned int repeat[32];
};
extern struct Input Inp;

int InputInit();
int InputUpdate(int EnableDiagnals);

// mdrom_good
extern int MD_DAT_LookFor(int CRC32);
extern char *MD_DAT_getname(int no);

// smsrom_good
extern int SMS_DAT_LookFor(int CRC32);
extern char *SMS_DAT_getname(int no);

// rom_crc
extern unsigned int CRC;
extern unsigned int CRC32_table[];
#define GetCRC(Byte) CRC=(CRC32_table[((CRC^0xffffffffL)^(Byte))&0xff]^((CRC^0xffffffffL))>>8)^0xffffffffL;

// render8.s
extern void draw_string_8(char *text,unsigned short x,unsigned short y,unsigned char *buffer);
extern void set_font_8(unsigned char *font);
extern void set_font_color_8(unsigned char low,unsigned char high);

// gp32_snc.c
extern unsigned int sound_buffer_size;

void flushcache(void);

#endif /* _APP_H_ */





