
#include "app.h"


/*
    Load a normal file, or ZIP/GZ archive.
    Returns NULL if an error occured.
*/
void get_archive_filename(char *filename, char *longfilename)
{
   FILE *fd = NULL;
   unsigned char romname_len=0;
   char buf[1];
   fd = (FILE*)fopen(filename, "rb");
   if(fd)
   {    
     fseek(fd, 26,SEEK_SET);
     if(fread(buf, 1, 1, fd)==1) // read 1 byte - name length
     {
	// now copy name string to te
	romname_len=(unsigned char)buf[0];
	longfilename[0]=romname_len;
	fseek(fd, 30,SEEK_SET);
	fread(longfilename,1,romname_len,fd);
	if(longfilename[romname_len-4] == '.')
	{
	  longfilename[romname_len-1]=0;
	  longfilename[romname_len-2]=0;
	  longfilename[romname_len-3]=0;
	  longfilename[romname_len-4]=0;
	}
     }
     fclose(fd);
    }
}
			
int get_archive_crc(char *filename)
{
   FILE *fd = NULL;
   unsigned char romname_len=0;
   int buf;
   
   fd = (FILE*)fopen(filename, "rb");
   if(fd)
   {    
     fseek(fd, 14,SEEK_SET);
     if(fread((char*)&buf, 1, 4, fd)==4) // read 1 byte - name length
     {
        fclose(fd);
	return(buf);
     }
     else
     {
       fclose(fd);
       return(0);
     }
    }
    
    return(0);
}

int load_archive(char *filename, char *buffer, int *file_size,int flip)
{
    int size = 0;
    uint8 *buf = NULL;
    char text[128];	
    if(check_zip(filename))
    {
        unzFile *fd = NULL;
        unz_file_info info;
        int ret = 0;
         
        /* Attempt to open the archive */
        fd = unzOpen(filename);
        if(!fd)
	{
	   gp_drawString(8,100,19,"Failed to open file",(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	   return (0);
	}

        /* Go to first file in archive */
        ret = unzGoToFirstFile(fd);
        if(ret != UNZ_OK)
        {
	    gp_drawString(8,100,32,"Failed to find first file in zip",(unsigned short)RGB(31,31,31),framebuffer16[flip]);
            unzClose(fd);
            return (0);
        }

        ret = unzGetCurrentFileInfo(fd, &info, NULL, 0, NULL, 0, NULL, 0);
        if(ret != UNZ_OK)
        {
	    sprintf(text,"Failed to zip info");
	    gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
            unzClose(fd);
            return (0);
        }

        /* Open the file for reading */
        ret = unzOpenCurrentFile(fd);
        if(ret != UNZ_OK)
        {
	    gp_drawString(8,100,19,"Failed to read file",(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	    sprintf(text,"err: %d",ret);
	    gp_drawString(8,110,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
            unzClose(fd);
            return (0);
        }

        /* Allocate file data buffer */
        size = info.uncompressed_size;
	if(size>ROM_SIZE)
	{
	   gp_drawString(8,100,21,"Rom exceeds 4MB limit",(unsigned short)RGB(31,31,31),framebuffer16[flip]);
           unzClose(fd);
           return (0);
	}
	buf = buffer;

        /* Read (decompress) the file */
        ret = unzReadCurrentFile(fd, buf, info.uncompressed_size);
        if(ret != info.uncompressed_size)
        {
            //free(buf);
	    gp_drawString(8,100,20,"File failed to uncompress fully",(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	    unzCloseCurrentFile(fd);
            unzClose(fd);
            return (0);
        }

        /* Close the current file */
        ret = unzCloseCurrentFile(fd);
        if(ret != UNZ_OK)
        {
            //free(buf);
	    sprintf(text,"Failed to close file in zip");
	    gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
            unzClose(fd);
            return (0);
        }

        /* Close the archive */
        ret = unzClose(fd);
        if(ret != UNZ_OK)
        {
            //free(buf);
	    sprintf(text,"Failed to close zip");
	    gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
            return (0);
        }

        /* Update file size and return pointer to file data */
        *file_size = size;
        return (1);
    }
    else
    {
        sprintf(text,"Not zip file");
	    gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	while(1==1)
	{
	}
        gzFile *gd = NULL;

        /* Open file */
        gd = gzopen(filename, "rb");
        if(!gd) return (0);

        /* Get file size */
        size = gzsize(gd);

        /* Allocate file data buffer */
	buf = RomData;

        /* Read file data */
        gzread(gd, buf, size);

        /* Close file */
        gzclose(gd);

        /* Update file size and return pointer to file data */
        *file_size = size;
        return (1);
    }
}


/*
    Verifies if a file is a ZIP archive or not.
    Returns: 1= ZIP archive, 0= not a ZIP archive
*/
int check_zip(char *filename)
{
    uint8 buf[2];
    FILE *fd = NULL;
    fd = (FILE*)fopen(filename, "rb");
    if(!fd) return (0);
    fread(buf, 1, 2, fd);
    fclose(fd);
    if(memcmp(buf, "PK", 2) == 0) return (1);
    return (0);
}


/*
    Returns the size of a GZ compressed file.
*/
int gzsize(gzFile *gd)
{
    #define CHUNKSIZE   (0x10000)
    int size = 0, length = 0;
    unsigned char buffer[CHUNKSIZE];
    gzrewind(gd);
    do {
        size = gzread(gd, buffer, CHUNKSIZE);
        if(size <= 0) break;
        length += size;
    } while (!gzeof(gd));
    gzrewind(gd);
    return (length);
    #undef CHUNKSIZE
}

int save_archive(char *filename, char *buffer, int size, int flip)
{
    uint8 *buf = NULL;
    char text[128]="";	
    zipFile *fd = NULL;
    int ret = 0;
    fd=zipOpen(filename,0);
    if(!fd)
    {
       sprintf(text,"Failed to create zip");
       gp_drawString(8,60,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
       return (0);
    }
	//sprintf(text,"created zip");
    //gp_drawString(0,10,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
    ret=zipOpenNewFileInZip(fd,"DRMDSAVE",
			    NULL,
                            NULL,0,
			    NULL,0,
			    NULL,
			    Z_DEFLATED,
			    Z_BEST_COMPRESSION);
			    
    if(ret != ZIP_OK)
    {
       zipClose(fd,NULL);
       sprintf(text,"Failed to create file in zip");
       gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
       return (0);    
    }
    //sprintf(text,"created file in zip");
    //gp_drawString(0,20,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	
    ret=zipWriteInFileInZip(fd,buffer,size);
    if(ret != ZIP_OK)
    {
      zipCloseFileInZip(fd);
      zipClose(fd,NULL);
      sprintf(text,"Failed to write file in zip");
      gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
      return (0);
    }
	//sprintf(text,"write file in zip");
    //gp_drawString(0,30,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	
    ret=zipCloseFileInZip(fd);
    if(ret != ZIP_OK)
    {
      zipClose(fd,NULL);
      sprintf(text,"Failed to close file in zip");
      gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
      return (0);
    }
	//sprintf(text,"close file in zip");
    //gp_drawString(0,40,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	
    ret=zipClose(fd,NULL);
    if(ret != ZIP_OK)
    {
      sprintf(text,"Failed to close zip");
      gp_drawString(8,100,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
      return (0);
    }
	
	//sprintf(text,"close zip");
    //gp_drawString(0,50,strlen(text),text,(unsigned short)RGB(31,31,31),framebuffer16[flip]);
	
    return(1);
}
